;;; -*- Mode:Lisp; Package:KEE; Base:10; Syntax: Common-lisp -*-
;;; ex: set lisp ai sm:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Copyright, 1990. The Regents of the University of California.
;;; This software was produced under a U. S. Government contract
;;; (W-7405-ENG-36) by Los Alamos National Laboratory, which is
;;; operated by the University of California for the U. S. Department
;;; of Energy.  The U. S. Government is licensed to use, reproduce,
;;; and distribute this software. Permission is granted to the
;;; public to copy and use this software without charge, provided
;;; that this Notice and any statement of authorship are reproduced
;;; on all copies.  Neither the Government nor the University makes
;;; any warranty, express or implied, or assumes any liability or
;;; responsibility for the use of this software.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The biggest limitation of this scheme is that I do not have true generic
;;; functions. Method combination is performed by KEE, and method invocation
;;; is via (unitmsg unit method args ...).
;;; I fake a generic function with a "real" defun'ed function. The function just
;;; turns into a unitmsg variant.
;;;
;;; As everywhere in this KEE re-tread, consistency (e.g. congruent lambda lists)
;;; is given short shrift.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'kee)

(eval-when (compile load eval)
  (defun process-lambda-list (lambda-list)
    "Strip &rest and &optional from defgeneric calls."
    (labels
	(;;
	 ;; strip-lambda-keywords removes all &optionals and any &rest from the
	 ;; lambda list so that it can be use in a unitmsg* call.
	 ;;
	 (strip-lambda-keywords (lambda-list)
	   (cond ((null lambda-list) ())
		 ;;
		 ;; (&rest foo)
		 ;; &rest is just removed entirely. The parameter is handeled
		 ;; elsewhere.
		 ;;
		 ((eql '&rest (car lambda-list))
		  ())
		 ;;
		 ;; (&optional ...)
		 ;;
		 ((eql '&optional (car lambda-list))
		  (strip-lambda-keywords (cdr lambda-list)))
		 ;;
		 ;; (foo ...)
		 ;;
		 ((atom (car lambda-list))
		  (cons (car lambda-list) (strip-lambda-keywords (cdr lambda-list))))
		 ;;
		 ;; ((foo default-value) ...) following an &optional
		 ;;
		 ((listp (car lambda-list))
		  (cons (caar lambda-list)
			(strip-lambda-keywords (cdr lambda-list))))))
	 ;;
	 ;; find-rest-arg looks for the (&rest rest-args) part and returns rest-args
	 ;;
	 (find-rest-arg (lambda-list)
	   (cond ((null lambda-list) ())
		 ((eql '&rest (car lambda-list)) (cadr lambda-list))
		 (t (find-rest-arg (cdr lambda-list)))))
	 )
      (let ((rest-arg (find-rest-arg lambda-list))
	    (basic-args (strip-lambda-keywords lambda-list)))
	(if rest-arg
	    (cond ((null basic-args) rest-arg)
		  ((eql 1 (length basic-args)) `(cons ,(car basic-args) ,rest-arg))
		  (t `(nconc (list ,@basic-args) ,rest-arg)))
	    (if (null basic-args)
		()
		`(list ,@basic-args)))))))

(defmacro defgeneric (function lambda-list &body forms)
  "The macro that will generate a defun for a set of methods."
  (let ((doc "")
	(decls ())
	(methods ()))
    ;;
    ;; Parse any options to defgeneric saving a doc string, declarations and methods.
    ;;
    (dolist (form forms)
      (case (car form)
	;;
	;; Unsupported options
	;;
	((:argument-precedence-order :method-combination :generic-function-class)
	 (error "Unsupported defgeneric option."))
	;;
	;; doc string
	;;
	(:documentation
	 (setf doc (second form)))
	;;
	;; declarations
	;;
	(declare
	 (if (null decls)
	     (setf decls (cdr form))
	     (nconc decls (cdr form))))
	;;
	;; methods
	;;
	(:method
	 (push (nconc (list 'defmethod function) (cdr form)) methods))
	;;
	;; Anything else is a typo
	;;
	(otherwise
	 (error "Unknown defgeneric option."))
	))
    ;;
    ;; Now build a defun with what we know. Allow the bundling of any defmethods.
    ;;
    (let ((body-form `(defun ,function ,lambda-list
		       ,doc
		       ,@(if (null decls)
			     ()
			     `((declare ,@decls)))
		       (unitmsg*
			,(car lambda-list)
			',function
			,(process-lambda-list (cdr lambda-list))))))
      (if (not (null methods))
	  `(progn ,body-form ,@methods)
	  body-form))))
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; defmethod is used to define kee methods.
;;; A major difference from true CLOS is that defmethod does not automatically
;;; define a generic function. This reflects:
;;;   1. My generic functions are just syntactic sugar for (unitmsg ...)
;;;   2. It is considered good programming practice to always do a defgeneric
;;;      before any defmethods anyway.
;;; The syntax is similar to CLOS defmethod except:
;;;
;;; 1. Specialized Parameters
;;;
;;; There is exactly one "specialized parameter" which must be the first
;;; parameter in the lambda list.
;;; This reflects the fact that:
;;;
;;; a. KEE methods are attached to individual units.
;;; b. Multiple inheritance is handled by a different mechanism than in CLOS.
;;; c. unitmsg always places the self-reference to its target as the first parameter
;;;    of the method.
;;;
;;; The "parameter-specializer-name" associated with this first parameter is the
;;; name of the object (unit) to which the method is attached. The method name
;;; is the slot to which the method will be attached.
;;;
;;; 2. No modification of setf methods
;;;
;;; The form (defmethod (setf foo) ...) is not supported.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The following forms are supported.
;;;
;;; (defmethod method-name ((self unit) arg ...) form1 ... formn)
;;;	will place a primary method in the slot "method-name" of unit "unit". The first
;;;	parameter to the method is "self" and the rest of the parameters are "arg..."
;;;	The remaining forms provide the body of the method. Note that unlike
;;;     CLOS primary methods where (call-next-method) can be used to get to
;;;     less specific primary methods, KEE wrapperbody's can only be used in
;;;     :around methods. Thus, nothing is done with (call-next-method)
;;;     forms here. They are only allowed in :around methods.
;;;
;;; (defmethod method-name :before ((self unit) arg ...) form1 ... formn)
;;; (defmethod method-name :after ((self unit) arg ...) form1 ... formn)
;;;	will add before and after wrappers to a method. The specialised parameter
;;;     is not really used because KEE before and after methods are evaluated within
;;;	the scope of the primary method's parameters. This means that the entire
;;;	parameter list for the before and after method need not be given. The only
;;;	use made here of the parameter list is to name the unit to which the
;;;	before or after wrapper is to be attached. This form was chosen, despite its
;;;	possibility for syntactic (and unchecked!) conflict with the primary method,
;;;	because it seemed best to have as similar a syntax as possible for both
;;;	primary and wrapper methods.
;;;
;;; (defmethod method-name :around ((self unit) arg ...) form1 ... formn)
;;;     will add a wrapper method to a method slot. The parameter list has the
;;;     same limitations as those mentioned above for :before and :after.
;;;     Each form1 ... formn is examined to see if it is the form
;;;     (call-next-method). If it is, it is replaced with a KEE wrapperbody.
;;;     Note that the use of call-next-method is limited to :around methods
;;;     rather than being usable in primary methods as in CLOS. Also, none
;;;     of the additional functionality of call-next-method, such as altering
;;;     the parameter list, is allowed here. This reflects the limitations (?)
;;;     of the KEE wrapperbody.
;;;     Also, (in the most blatant differentation from the CLOS standard)
;;;     the return value from call-next-method cannot be used due to the nature of
;;;     the KEE wrapperbody.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro defmethod (&body args)
  "Programatically define KEE methods as forms in KEE method slots."
  (labels
      ;;
      ;; change is used by the :around method code to change (call-next-method)
      ;; into wrapperbody. The body form is copied.
      ;;
      ((change (body-form)
	 (cond ((not (consp body-form))
		body-form)
	       ((and (eql (car body-form) 'call-next-method)
		     (eql (length body-form) 1))
		'wrapperbody)
	       (t (cons (change (car body-form))
			(change (cdr body-form))))))
       ;;
       ;; build does all the work of generating the KEE code for all the
       ;; different cases. The main part of the macro just makes sure that
       ;; build is called with a correct and consistent set of parameters.
       ;;
       (build (type slot arg-list &rest body)
	 (let ((self-ref (car arg-list)))
	   ;;
	   ;; self-ref is of the form (parameter-name object-name)
	   ;; and works like a CLOS argument. parameter-name will be
	   ;; something like self, and object-name will be the unit
	   ;; to which we add the local contribution to the method
	   ;; defined by this defmethod.
	   ;;
	   (if (not (and (listp self-ref)
			 (= (length self-ref) 2)))
	       (error "first parameter to method not of the form (name object)."))
	   ;;
	   ;; The final form consists of a common body of error-checking code
	   ;; that creates the units and slots as needed.
	   ;; This common body then has the proper add.value appended based on the
	   ;; method type.
	   ;;
	   ;; args are the parameter list to be used by the lambda representing
	   ;;	    the method.
	   ;; unit is the target object.
	   ;; slot-comment is the string (possibly nil) to be used in the comment
	   ;;       facet of the slot that contains this method.
	   ;;	    This comment is found by looking at the forms that make up
	   ;        the body. If the first form is a string,
	   ;;	    and if there is more than one form, then this string is the
	   ;;	    documentation string for the method. This usage parallels
	   ;;	    the documentation string of defuns and lambdas.
	   ;; body-forms are the actual forms that make up the body with the
	   ;;	    documentation string removed.
	   ;;
	   (let ((args (cons (car self-ref) (cdr arg-list)))
		 (unit (cadr self-ref))
		 (slot-comment (if (and (> (length body) 1)
					(stringp (car body)))
				   (car body)
				   nil))
		 (body-forms (if (and (> (length body) 1)
				      (stringp (car body)))
				 (cdr body)
				 body)))
	     (append
	      ;;
	      ;; error-checking for existence of proper units and slots
	      ;;
	      `(progn 
		(if (not (unit.exists.p ',unit))
		    (error "defmethod attempts to add a method to a non-existent object"))
		(if (not (slot.exists.p ',slot
			      ',unit
			      ()
			      'member))
		    (create.slot ',unit
			   ',slot
			   'member
			   ()
			   'method
			   'method)))
	      ;;
	      ;; creation or removal of the slot comment based on the
	      ;; existence of a documentation string
	      ;;
	      (if (null slot-comment)
		  `((remove.all.facet.values ',unit
		     ',slot
		     'comment))
		  `((put.facet.value ',unit
		     ',slot
		     'comment
		     ,slot-comment)))
	      ;;
	      ;; The actual method code
	      ;;
	      (ecase type
		(:primary
		 `((add.value ',unit
		    ',slot
		    '(lambda ,args ,@body-forms)
		    'member)))
		(:around
		 `((add.value ',unit
		    ',slot
		    '(wrapper (progn ,@(change body-forms)))
		    'member)))
		(:after
		 `((add.value ',unit
		    ',slot
		    '(after ,@body-forms)
		    'member)))
		(:before
		 `((add.value ',unit
		    ',slot
		    '(before ,@body-forms)
		    'member)))))))))
    ;;
    ;; This is the main body of the macro. Its only purpose is to call the build
    ;; routine with correct parameters removing :before and :after.
    ;;
    (append
     (case (cadr args)
       (:after
	(apply #'build (cons :after (cons (car args) (cddr args)))))
       (:before
	(apply #'build (cons :before (cons (car args) (cddr args)))))
       (:around
	(apply #'build (cons :around (cons (car args) (cddr args)))))
       (otherwise
	(apply #'build (cons :primary args))))
     (case (cadr args)
       (:after
	(list `',(read-from-string (format nil
					   "|~s>~s<AFTER>|"
					   (second (first (third args)))
					   (car args)))))
       (:before
	(list `',(read-from-string (format nil
					   "|~s>~s<BEFORE>|"
					   (second (first (third args)))
					   (car args)))))
       (:around
	(list `',(read-from-string (format nil
					   "|~s>~s<AROUND>|"
					   (second (first (third args)))
					   (car args)))))
       (otherwise
	(list `',(read-from-string (format nil
					   "~s>~s"
					   (second (first (second args)))
					   (car args)))))))))
